(* This file belongs to the Secure Sessions Project 	      *)
(* http://www.msr-inria.inria.fr/projects/sec/sessions/       *)
(* MSR-INRIA Joint Centre, 2007    			      *)

module Crypto

open System.Security.Cryptography
open Printf
open Data
open Global
open Base64

type keybytes = 
    SymKey of bytes
  | RSAKey of RSACryptoServiceProvider
(*
  | Params of RSAParameters
  | XmlString of string
*)
  
(* random number generation *) 

let rng = new RNGCryptoServiceProvider ()
let random () : str = 
  let x = Bytearray.make 16 in
  rng.GetBytes x; base64 x

let mkNonce () : bytes =
  let x = Bytearray.make 16 in
  rng.GetBytes x; x

(*---- Hashing. *)

let sha1_instance = SHA1.Create ()
let sha1 (x:bytes) : bytes = sha1_instance.ComputeHash x

let hmacsha1 (SymKey key) (x:bytes) : bytes = (new HMACSHA1 (key)).ComputeHash x

(*---- X509. *)

let some_x509 =
  ibase64 
"MIIBxDCCAW6gAwIBAgIQxUSXFzWJYYtOZnmmuOMKkjANBgkqhkiG9w0BAQQFADAW
MRQwEgYDVQQDEwtSb290IEFnZW5jeTAeFw0wMzA3MDgxODQ3NTlaFw0zOTEyMzEy
MzU5NTlaMB8xHTAbBgNVBAMTFFdTRTJRdWlja1N0YXJ0Q2xpZW50MIGfMA0GCSqG
SIb3DQEBAQUAA4GNADCBiQKBgQC+L6aB9x928noY4+0QBsXnxkQE4quJl7c3PUPd
Vu7k9A02hRG481XIfWhrDY5i7OEB7KGW7qFJotLLeMec/UkKUwCgv3VvJrs2nE9x
O3SSWIdNzADukYh+Cxt+FUU6tUkDeqg7dqwivOXhuOTRyOI3HqbWTbumaLdc8juf
z2LhaQIDAQABo0swSTBHBgNVHQEEQDA+gBAS5AktBh0dTwCNYSHcFmRjoRgwFjEU
MBIGA1UEAxMLUm9vdCBBZ2VuY3mCEAY3bACqAGSKEc+41KpcNfQwDQYJKoZIhvcN
AQEEBQADQQAfIbnMPVYkNNfX1tG1F+qfLhHwJdfDUZuPyRPucWF5qkh6sSdWVBY5
sT/txBnVJGziyO8DPYdu2fPMER8ajJfl" 

// c:\Program Files\Microsoft WSE\v2.0\Samples\Sample Test Certificates\Client Private.pfx
// openssl pkcs12 -in key.pfx -password pass:wse2qs
// openssl rsa -text -in key.pem -passin pass:hello
let x509privkey : bytes = ibase64 
"MIICXQIBAAKBgQC+L6aB9x928noY4+0QBsXnxkQE4quJl7c3PUPdVu7k9A02hRG4
81XIfWhrDY5i7OEB7KGW7qFJotLLeMec/UkKUwCgv3VvJrs2nE9xO3SSWIdNzADu
kYh+Cxt+FUU6tUkDeqg7dqwivOXhuOTRyOI3HqbWTbumaLdc8jufz2LhaQIDAQAB
AoGAIbKXshdzP8Qe1iIsctaAYlzC2IrBEhQLpoH4cFNi6LZFUQ+q4DZdULTHt5Aj
VmvaQlkHGJMXiNCMwPiZhbtrIVb6EaC+3leFQntQH9989uBOOvnGOYK8TYR7fy8g
p2SnIrbwnK81xMVHOZ+ip4F9JjCoJFu7JGQ8LJys4clFfMECQQDdm0DHgEkFTqnX
6gXRYAvzT5ymNnLg4uoeufUDdZfpIF98sHWeyin7SaNRvZeOUjfWKZMhHfcxDujz
BWrnyb2VAkEA27QFsZNIvAU91EzVgk//W1EETj3BmWoMMJ0RQcl+4KZn5H9SJiL2
1lV0qV+vPyI240pE8NZ4/2I9TcBWdJ8XhQJBAMz4UaV1QZT4Gl5L5c+wH72PsFuI
Vx8sotGCvF3TwYtuVwOUj2YspUH47Weojeqkn2KmF6D0/3NP0CUfqK71STUCQQDY
UTzV5Y1UythgwrHqF3f2YiOaIcYPllGgP9Xw70LEUDO00AODVyOwTv2JkpOK1tYy
HjsY4iLKWODBQCIfu1C9AkBS+s04OsH3jJRGc74KsAhqC7r056rb8oOOc+6X3K4N
nB3YAlyMvZBK3I8BorugTnrxANak1kZPm79LZJ1r3M/7"

let x509signcert (i:str) (s:str) (key:bytes) :bytes = failwith "cannot generate X.509 certs"

let ix509 (x:bytes) : string * string * bytes =
  let c = new System.Security.Cryptography.X509Certificates.X509Certificate (x) in
  (c.Subject, c.GetKeyAlgorithm (), c.GetPublicKey ())
let x509key (x:bytes) : bytes = 
  let c = new System.Security.Cryptography.X509Certificates.X509Certificate (x) in
  c.GetPublicKey()
let x509issuer (x:bytes) : str = 
  let c = new System.Security.Cryptography.X509Certificates.X509Certificate (x) in
  c.get_Issuer()

let x509subject (x:bytes) : str = 
  let c = new System.Security.Cryptography.X509Certificates.X509Certificate (x) in
  c.get_Subject()

let x509number (x:bytes) : str = 
  let c = new System.Security.Cryptography.X509Certificates.X509Certificate (x) in
  c.GetSerialNumberString()

let x509alg (x:bytes) : str = 
  let c = new System.Security.Cryptography.X509Certificates.X509Certificate (x) in
  c.GetKeyAlgorithm()



// fixme: currently x509 only works with the hard-wired get_prikey and x509 certificate above
let get_auth_prikey () = [||]
let get_auth_pubkey () = [||]
let get_pubkey () = [||]

let get_user () : string = "t-stetse"
let get_pass () : string = base64 (rev (utf8 (get_user ())))


(*---- RSA. *)

let rsa_instance = new RSACryptoServiceProvider ()
(*
let rsa_sign (key:bytes) (hash:bytes) : bytes =
  let mutable rkey = new RSAParameters() in
  rkey.Modulus <- Array.sub key 11 128;
  rkey.Exponent <- Array.sub key 141 3;
  rkey.D <- Array.sub key 147 128;
  rkey.P <- Array.sub key 278 64;
  rkey.Q <- Array.sub key 345 64;
  rkey.DP <- Array.sub key 412 64;
  rkey.DQ <- Array.sub key 479 64;
  rkey.InverseQ <- Array.sub key 545 64;
  rsa_instance.ImportParameters rkey;
  let oid = CryptoConfig.MapNameToOID "sha1" in
  rsa_instance.SignHash (hash, oid)

let rsa_verify (key:bytes) (hash:bytes) (x:bytes) : unit =
  let mutable rkey = new RSAParameters() in
  rkey.Exponent <- Array.sub key (key.Length - 3) 3;
  rkey.Modulus <- Array.sub key (key.Length - 128 - 5) 128;
  rsa_instance.ImportParameters rkey;
  let oid = CryptoConfig.MapNameToOID "sha1" in
  if not (rsa_instance.VerifyHash (hash, oid, x))
  then failwith "RSA_verify failed"
*)

//RSA Encryption

(*
let rsa_keygen () = 
  let rsa = new RSACryptoServiceProvider() 
  let rkey = rsa.ExportParameters(true) in 
    Params(rkey)
*)

let rsa_keygen () = 
  let rsa = new RSACryptoServiceProvider() in
    RSAKey rsa

(*
let rsa_pub (RSA skey) = 
  let mutable pub_key = new RSAParameters() in
  pub_key.Exponent <- rkey.Exponent;
  pub_key.Modulus <- rkey.Modulus;
  Params pub_key
*)

let rsa_pub (RSAKey skey) = 
  let pub = skey.ExportParameters(false) in
  let rsa = new RSACryptoServiceProvider () in
    rsa.ImportParameters(pub);
    RSAKey rsa

(*    
let rsa_encrypt (Params pkey) (v:bytes)  = 
  rsa_instance.ImportParameters(pkey);
  rsa_instance.Encrypt(v,false)
*)
    
let rsa_encrypt (RSAKey pkey) (v:bytes)  = 
  pkey.Encrypt(v,false) (* setting OAEP = true for indigo, it is false for WSE *)

let rsa_encrypt_oaep (RSAKey pkey) (v:bytes)  = 
  pkey.Encrypt(v,true) (* setting OAEP = true for indigo, it is false for WSE *)


(*
let rsa_decrypt (Params skey) (v:bytes) =
  rsa_instance.ImportParameters(skey);
  rsa_instance.Decrypt(v,false)
*)

let rsa_decrypt (RSAKey skey) (v:bytes) =
  skey.Decrypt(v,false) (* setting OAEP = true for indigo, it is false for WSE *)
let rsa_decrypt_oaep (RSAKey skey) (v:bytes) =
  skey.Decrypt(v,true) (* setting OAEP = true for indigo, it is false for WSE *)

(*
let rsa_sign (Params pkey) (v:bytes)  = 
  rsa_instance.ImportParameters(pkey);
  let oid = CryptoConfig.MapNameToOID "sha1" in
  rsa_instance.SignHash (sha1 v, oid)
*)

let rsa_sign (RSAKey skey) (v:bytes)  = 
  let oid = CryptoConfig.MapNameToOID "sha1" in
  skey.SignHash (sha1 v, oid)

(*
let rsa_verify (Params skey) (v:bytes) (x:bytes)  =
  rsa_instance.ImportParameters(skey);
  let oid = CryptoConfig.MapNameToOID "sha1" in
  if not (rsa_instance.VerifyHash (sha1 v, oid, x))
  then failwith "RSA_verify failed"
*)

let rsa_verify (RSAKey pkey) (v:bytes) (x:bytes)  =
  let oid = CryptoConfig.MapNameToOID "sha1" in
  if not (pkey.VerifyHash (sha1 v, oid, x))
  then failwith "RSA_verify failed"

let rsa_skey (key:str) = 
  let rsa = new RSACryptoServiceProvider () in
    rsa.FromXmlString(key);
    RSAKey rsa

let rsa_pkey (key:str) = rsa_pub (rsa_skey key)

let rsa_pkey_bytes (key:byte[]) = 
  let mutable rkey = new RSAParameters() in
  rkey.Exponent <- Array.sub key (key.Length - 3) 3;
  rkey.Modulus <- Array.sub key (key.Length - 128 - 5) 128; 
(*  rkey.Modulus <- Array.sub key 11 128;
  rkey.Exponent <- Array.sub key 141 3;
  rkey.D <- Array.sub key 147 128;
  rkey.P <- Array.sub key 278 64;
  rkey.Q <- Array.sub key 345 64;
  rkey.DP <- Array.sub key 412 64;
  rkey.DQ <- Array.sub key 479 64;
  rkey.InverseQ <- Array.sub key 545 64; *)
  let rsa = new RSACryptoServiceProvider () in
    rsa.ImportParameters(rkey);
    RSAKey rsa

let rsa_skey_bytes (key:byte[]) = 
  let mutable rkey = new RSAParameters() in
  rkey.Modulus <- Array.sub key 11 128;
  rkey.Exponent <- Array.sub key 141 3;
  rkey.D <- Array.sub key 147 128;
  rkey.P <- Array.sub key 278 64;
  rkey.Q <- Array.sub key 345 64;
  rkey.DP <- Array.sub key 412 64;
  rkey.DQ <- Array.sub key 479 64;
  rkey.InverseQ <- Array.sub key 545 64; 
  let rsa = new RSACryptoServiceProvider () in
    rsa.ImportParameters(rkey);
    RSAKey rsa

let x509_encrypt (key:bytes) (v:byte[]) = 
  let rkey = rsa_pkey_bytes key in
    rsa_encrypt rkey v
let x509_decrypt (key:bytes) (v:byte[]) = 
  let rkey = rsa_skey_bytes key in
    rsa_decrypt rkey v

let random_bytes(n:int): bytes = 
  let x = Bytearray.make n in
    rng.GetBytes x; x

(*---- AES. *)

(* random_bytes 16 ??*)
let mkKey () = (mkNonce()) // todo: check length 

(*let mkKey256 () : bytes = (mkNonce256())*)

let mkPassword () = base64 (mkNonce()) // todo: check length 



(* AES is block-based, so does implicit padding. To make decryption an inverse of
   encryption, we use concatvar, so as to prepend the length of the message. This
   is a waste if we know what the length is, but concatenating empty lists is
   symbolically tricky. We may provide a "raw AES" primitive if needed. *)
   
   
let pad plain = concat plain [| |]
let unpad plain = let v, pad = iconcat plain in v 

let aes_encrypt (SymKey key) (v:bytes) = 
  let plain = v in
  let keyl = 8 * key.Length  in
  let block = 16 in (* 128 bits *)
  let rij = new RijndaelManaged() in
  rij.set_KeySize(keyl);
//  rij.set_Padding(PaddingMode.ISO10126);
  rij.GenerateIV();
//let iv = random_bytes (block) in
  let iv = rij.get_IV() in
  let enc = rij.CreateEncryptor(key,iv) in
  let mems = new System.IO.MemoryStream() in
  let crs = new CryptoStream(mems,enc,CryptoStreamMode.Write) in
  let _ =  crs.Write(plain,0,plain.Length) in
  let _ = crs.FlushFinalBlock() in
  let cip = mems.ToArray() in
  let ctxt = Bytearray.make (iv.Length + cip.Length) in
  let _ = System.Array.Copy(iv,0,ctxt,0,iv.Length) in
  let _ = System.Array.Copy(cip,0,ctxt,iv.Length,cip.Length) in
    mems.Close();
    crs.Close();
    ctxt

let aes_decrypt (SymKey key) (cipher:bytes) = 
  let rij = new RijndaelManaged() in
  let keyl = 8 * key.Length  in
  rij.set_KeySize(keyl);
  let block = 16 in (* 128 bits *)
  let iv = Bytearray.make (block) in
  let _ = System.Array.Copy(cipher,0,iv,0,iv.Length) in
  let cip = Bytearray.make (cipher.Length - iv.Length) in
  let _ = System.Array.Copy(cipher,iv.Length,cip,0,(cipher.Length - iv.Length)) in
  let dec = rij.CreateDecryptor(key,iv) in
  let mems = new System.IO.MemoryStream(cip) in
  let crs = new CryptoStream(mems,dec,CryptoStreamMode.Read) in
  let plain = Bytearray.make(cipher.Length) in  
  let _ =  crs.Read(plain,0,plain.Length) in
    plain




let keytobytes kb = match kb with
    SymKey(k) -> k
  | RSAKey(k) -> failwith "keytobytes: asym not implemented concretely"

let symkey k = SymKey k

let asympubkey k = (rsa_pkey_bytes k)
let asymprivkey k = (rsa_skey_bytes k)

let asympub k = (rsa_pkey k)
let asympriv k = (rsa_skey k)

