(* This file belongs to the Secure Sessions Project 	      *)
(* http://www.msr-inria.inria.fr/projects/sec/sessions/       *)
(* MSR-INRIA Joint Centre, 2007    			      *)

{
  (********************************************************************)
  (* Session - Implementation                                          *)
  (*                                                                  *)
  (* lexer.mll                                                        *)
  (********************************************************************)
  (* $Id: lexer.mll 2847 2007-06-25 14:40:15Z denielou $ *)

open Common
open Syntax
open Parser

module LE = Lexing

let debug = Common.debug Common.debug_p "lexer"

let lexeme = LE.lexeme
let linestart = ref 0
let lineno = ref 1

let newline lexbuf : unit =
(*  debug "New line added" ; *)
  linestart := LE.lexeme_start lexbuf;
  incr lineno

let info lexbuf : info =
(*  debug "Gathering lexeme information" ; *)
  let c1 = LE.lexeme_start lexbuf in
  let c2 = LE.lexeme_end lexbuf in
  (!lineno, c1 - !linestart),(!lineno, c2 - !linestart)

let error lexbuf s =
  debug ("Error generation: "^s) ;
  let (l1,c1),(l2,c2) = info lexbuf in
  let t = lexeme lexbuf in
  let s = Printf.sprintf "%d:%d-%d: %s" l1 c1 c2 s in
  if t = "" then (debug ("No lexeme found") ;raise (Syntax_error (s,((l1,c1),(l2,c2)))))
  else (debug ("Lexeme found: "^t) ;raise (Syntax_error (s^ ": " ^ t,((l1,c1),(l2,c2)))))

}

let blank = [' ' '\t']+
let symbol = [^ '"' '=' '{' '}' '[' ']' '(' ')' ':'
                  '<' '>' ';' ',' '\n' '\r' ' ' '\t' '-' '+' '#' '$']+
let identifier = [ 'a' - 'z' 'A' - 'Z'] ['a' - 'z' 'A' - 'Z' '0' - '9']*
let ocamltype = [ 'a' - 'z' 'A' - 'Z' '0' - '9' '.' '-' '>' '*']+
let newline = '\n' 
let char = [^ '\n'] 


rule token = parse
  | blank         { token lexbuf }
  | "mu"          { MU  (info lexbuf) }
  | "role"        { ROLE (info lexbuf) }
  | "session"     { SESSION (info lexbuf) }
  | "="           { EQUAL (info lexbuf) }
  | "!"           { SEND (info lexbuf) }
  | "?"           { RECE (info lexbuf) }
  | ":"           { ocamltyp lexbuf }
  | ";"           { SEMI (info lexbuf) }
  | "."           { DOT (info lexbuf) }
  | "("           { LPA (info lexbuf) }
  | ")"           { RPA (info lexbuf) }
  | "<<"          { LANG (info lexbuf) }
  | ">>"          { RANG (info lexbuf,code lexbuf) }
  | "+"           { OR (info lexbuf) }
  | identifier    { IDENT (info lexbuf,lexeme lexbuf) }
  | eof		  { EOF (info lexbuf) }
  | "\r\n"        { newline lexbuf; token lexbuf }
  | newline       { newline lexbuf; token lexbuf }
  | "(*"          { comment lexbuf; token lexbuf }
  | _		  { error lexbuf "Unknown token" }

and comment = parse
  | "(*"             { comment lexbuf; comment lexbuf }
  | "*)"             { () }
  | newline          { newline lexbuf; comment lexbuf }
  | eof		     { error lexbuf "Unmatched '(*'" }
  | _                { comment lexbuf }

and code = parse
  | newline          { newline lexbuf; "\n"^(code lexbuf) }
  | eof		     { "" }
  | _ as x           { (String.make 1 x)^(code lexbuf) }

and ocamltyp = parse
  | ocamltype     { TYP (info lexbuf,lexeme lexbuf) }
  | _             { error lexbuf "Wrong type expression" }
