(* This file belongs to the Secure Sessions Project 	      *)
(* http://www.msr-inria.inria.fr/projects/sec/sessions/       *)
(* MSR-INRIA Joint Centre, 2007    			      *)

(********************************************************************)
(* Session - Implementation                                         *)
(*                                                                  *)
(* syntax.ml                                                        *)
(********************************************************************)
(* $Id: syntax.ml 2847 2007-06-25 14:40:15Z denielou $ *)

(*#use "c:/Documents and Settings/t-ricarc/inria-msr-sec/lang-sec/session-types/src/common.ml";;*)
(*#use "/Users/ricardocorin/inria-msr-sec/lang-sec/session-types/src/common.ml";;*)

open Common

(************************************)
(* Abstract syntax for source files *)
(************************************)

type principal = string

type message = string * string (* name of the constructor and the type of the sent value *)

type as_role_type = 
  | ASEnd
  | ASGoto of info * string
  | ASMu of info * string * as_role_type
  | ASSend of info * ((message * as_role_type) list)
  | ASReceive of info * ((message * as_role_type) list)

type as_session =
    (string * string * as_role_type) list

type ast = string * as_session * string

type node = int (* index *)


let node_ident = 
  Hashtbl.create 10
    
let clear_ident = 
  Hashtbl.clear node_ident

let add_ident (i:string) (n:node)=
  Hashtbl.add node_ident i n
    
let find_node (i:string) =
  Hashtbl.find node_ident i
    
let remove_ident (i:string) =
  Hashtbl.remove node_ident i


let nref = ref 0

let create_node_name ()=
  let i = !nref in 
  let _ = incr nref in
  i

let rec print_as_role  = function
  | ASEnd -> print_string "End "
  | ASGoto _ -> print_string "Goto "
  | ASMu (_,_,r) -> print_string "Mu " ; (print_as_role r)
  | ASSend (_,l) -> 
      begin
        print_string "Send " ; 
        List.iter (function (_,r) -> print_as_role r) l
      end
  | ASReceive (_,l) -> 
      begin
        print_string "Receive " ; 
        List.iter (function (_,r) -> print_as_role r) l
      end

    (*
let rec role_to_string  = function
  | ASEnd -> "End"
  | ASGoto _ -> "Goto"
  | ASMu (_,_,r) -> "Mu "^(role_to_string r)
  | ASSend (_,l) -> "Send"^()
  | ASReceive _ -> "Receive"
    *)



(*********************************************)
(* Graph-like representation of the protocol *)
(*********************************************)

type role_type =
    | End of node
    | Goto of node
    | Send of node * ((message * role_type) list)
    | Receive of node * ((message * role_type) list)

type session = (string*string*role_type) list


(* Conversion function from AS to role_type *)
let rec conversion (next:node option) = function
  | ASEnd -> 
      let n = create_node_name () in End n
  | ASGoto (_,i) -> Goto (find_node i)
  | ASMu (_,i,r) -> 
      let n = create_node_name () in
      let _ = add_ident i n in
      conversion (Some n) r
  | ASSend (_,l) -> 
      let n = match next with None -> create_node_name () | Some k -> k in
      Send (n, List.map (function (m,r) -> (m,conversion None r)) l)
  | ASReceive (_,l) -> 
      let n = match next with None -> create_node_name () | Some k -> k in
      Receive (n, List.map (function (m,r) -> (m,conversion None r)) l)

let to_role_type : as_session -> session =
  List.map 
    (function (x,y,z) -> 
      (* let _ = nref:=0 in *)
       (x,y,conversion None z))



(* Get the starting node of a role_type *)
let node_of : role_type -> node = function
  | End n -> n
  | Goto n -> n
  | Send (n,_) -> n
  | Receive (n,_) -> n


let rec sprint_role = function
  | End n -> Printf.sprintf "End %d " n
  | Goto n -> Printf.sprintf "Goto %d " n
  | Send (n,l) -> (Printf.sprintf "%d:Send{%s" n (sprint_list_messages l))
  | Receive (n,l) -> (Printf.sprintf "%d:Receive{%s" n (sprint_list_messages l))
and sprint_list_messages = function
    [] -> ""
  | ((c,t),r)::q -> (Printf.sprintf "|%s(%s).%s%s" c t (sprint_role r) (sprint_list_messages q))

let rec sprint_session :session -> string = function
    [] -> ""
  | (s,t,r)::q -> 
      begin
        s^" : "^t^"\n"^
        (sprint_role r)^"\n"^
        (sprint_session q)
      end

(************)
(* Examples *)
(************)

(* Correct example *)
let pingpong : session = 
  [("client","string",Send(0,[("Request","string"),
	   Receive(1,[(("Ack","unit"),Goto 0);
		      (("NAck","unit"),End 2)])]));
   ("server","string",Receive(3,[("Request","string"),
	   Send(4,[(("Ack","unit"),Goto 0);
		      (("NAck","unit"),End 5)])]))
  ]

(* Correct example *)
let ping : session = 
  [("client","string",Send(0,[("Request","string"),End 1]));
   ("server","string",Receive(2,[("Request","string"),End 3]))
  ]

(* Wrong example *)
let pingpong2 : session = 
  ["client","string",Send(0,[("Request","string"),
	   Receive(1,[(("Ack","unit"),Goto 0)])]);

   "server","string",Receive(2,[("Request","string"),
	   Send(3,[(("Ack","unit"),Goto 2);
		      (("NAck","unit"),End 4)])])
  ]

(* Wrong example *)
let pingpong2p : session = 
  ["client","string",Receive(2,[("Request","string"),
	   Send(3,[(("Ack","unit"),Goto 2);
		      (("NAck","unit"),End 4)])]);
   "server","string",Send(0,[("Request","string"),
	   Receive(1,[(("Ack","unit"),Goto 0)])])
  ]

(* Wrong example *)
let pingpong3 : session = 
  ["client","string",Send(0,[("Request","string"),
	   Receive(1,[(("Ack","unit"),Goto 0);
		      (("NAck","unit"),End 4)])]);
   "server","string",Receive(2,[("Request","string"),
	   Send(3,[(("Ack","unit"),End 5);
		      (("NAck","unit"),End 6)])])
  ]


let pingpongpang : session = 
  ["client","string", Send (15, [("M1","unit"), 
	                        Receive(16,[(("Last","unit"),End 10);
			                    (("Lastp","unit"),End 11)])]);
   
    "server","string",Receive(2,[("M1","unit"),
	                        Send(3,[(("M2","unit"),
		                        Receive(4,[("M3","unit"),
				                  Send(5,[(("Last","unit"),End 8)])]));
			                (("M2p","unit"),
		                        Receive(6,[("M3p","unit"),
				                  Send(7,[(("Lastp","unit"),End 9)])]))
		                       ])]);
    
    "bank","string",Receive(17,[(("M2","unit"),
		                Send(18,[(("M3","unit"),End 12)]));
		                (("M2p","unit"),
		                Send(19,[(("M3p","unit"),End 13)]))])
  ]
    
